use crate::vertex_2d::{ColoredVertex2D, Vertex2D};
use vulkano::command_buffer::{AutoCommandBufferBuilder, DynamicState};

use std::collections::HashMap;
use vulkano::buffer::{BufferAccess, BufferUsage, ImmutableBuffer};
use std::sync::Arc;
use vulkano::format::{ClearValue, Format};
use vulkano::framebuffer::{FramebufferAbstract, Framebuffer};
use vulkano::device::{Device, Queue};
use vulkano::instance::PhysicalDevice;
use vulkano::image::immutable::ImmutableImage;
use crate::util::shader_kernels::ShaderKernels;
use vulkano::image::{Dimensions, ImageAccess, ImageDimensions, SwapchainImage};
use vulkano::sampler::{Sampler, SamplerAddressMode, MipmapMode, Filter};
use vulkano::descriptor::DescriptorSet;
use vulkano::descriptor::descriptor_set::PersistentDescriptorSet;
use std::path::PathBuf;
use image::GenericImageView;
use crate::util::compute_image::ComputeImage;
use std::iter::FromIterator;
use vulkano::swapchain::Capabilities;

use winit::Window;
use vulkano::pipeline::viewport::Viewport;

// Canvas is the accumulator of Sprites for drawing

// Needs to know:
//   textured?
//   colored?
//   vertices

/*

If it is textured. It needs to be rendered with the texture shader which requires a separate
    graphics pipeline. Might as well have a new render pass as well.


So framebuffer is tied to the swapchains images as well as the renderpass

it appears that renderpass is tied to the individual shader


*/

// I want to be able to draw 2d sprites.

// These sprites might be textured or a single color

// All of the single colors will be grouped into one batch using colored vertices.
// The rest will be grouped by their texture and run individually


pub trait Vertex {
    fn position(&self) -> (f32, f32) {
        (0.0,0.0)
    }

    fn color(&self) -> Option<(f32, f32, f32, f32)> {
        Some((0.,0.,0.,0.))
    }

    fn textured(&self) -> bool {
        false
    }
}


impl Vertex for ColoredVertex2D {

    fn position(&self) -> (f32, f32) {
        (0.0,0.0)
    }

    fn color(&self) -> Option<(f32, f32, f32, f32)> {
        Some((0.,0.,0.,0.))
    }

    fn textured(&self) -> bool {
        false
    }
}

pub trait Drawable {
    fn get_vertices(&self) -> Vec<(f32, f32)>;
    fn get_color(&self) -> (f32, f32, f32, f32);
    fn get_texture_id(&self) -> Option<String>;
}



// Need three types of shaders. Solid, Textured, Compute
#[derive(PartialEq, Eq, Hash, Clone)]
pub enum ShaderType {
    SOLID = 0,
    TEXTURED = 1,
    COMPUTE = 2
}

#[derive(Clone)]
pub struct Canvas {

    colored_drawables : Vec<ColoredVertex2D>,
    colored_vertex_buffer: Vec<Arc<(dyn BufferAccess + std::marker::Send + std::marker::Sync + 'static)>>,

    textured_drawables: HashMap<String, Vec<Vertex2D>>,
    textured_vertex_buffer: HashMap<String, Arc<(dyn BufferAccess + std::marker::Send + std::marker::Sync + 'static)>>,

    shader_kernels: HashMap<ShaderType, ShaderKernels>,

    texture_store: HashMap<String, Arc<ImmutableImage<Format>>>,

    // Looks like we gotta hold onto the queue for managing textures
    queue: Arc<Queue>,
    sampler: Arc<Sampler>
}


impl Canvas {
    // needs to take in the texture list
    pub fn new(queue: Arc<Queue>,
               device: Arc<Device>,
               physical: PhysicalDevice,
               capabilities: Capabilities) -> Canvas {

        let solid_color_kernel = String::from("color-passthrough");
        let texture_kernel = String::from("simple_texture");

         let shader_kernels = HashMap::from_iter(vec![
             (ShaderType::SOLID, ShaderKernels::new(solid_color_kernel, capabilities.clone(), queue.clone(), physical.clone(), device.clone())),
             (ShaderType::TEXTURED, ShaderKernels::new(texture_kernel, capabilities.clone(), queue.clone(), physical.clone(), device.clone()))
         ]);

        Canvas {
            colored_drawables: vec![],
            colored_vertex_buffer: vec![],

            textured_drawables: Default::default(),
            textured_vertex_buffer: Default::default(),

            shader_kernels: shader_kernels,
            texture_store: Default::default(),

            queue: queue.clone(),
            sampler: Sampler::new(device.clone(), Filter::Linear, Filter::Linear,
                                MipmapMode::Nearest, SamplerAddressMode::Repeat, SamplerAddressMode::Repeat,
                                SamplerAddressMode::Repeat, 0.0, 1.0, 0.0, 0.0).unwrap(),

    }
    }

    // TODO Handle file not found gracefully
    fn get_texture_from_file(&self, image_filename: String) -> Arc<ImmutableImage<Format>> {

        let project_root =
            std::env::current_dir()
                .expect("failed to get root directory");

        let mut compute_path = project_root.clone();
        compute_path.push(PathBuf::from("resources/images/"));
        compute_path.push(PathBuf::from(image_filename));

        let img = image::open(compute_path).expect("Couldn't find image");

        let xy = img.dimensions();

        let data_length = xy.0 * xy.1 * 4;
        let pixel_count = img.raw_pixels().len();

        let mut image_buffer = Vec::new();

        if pixel_count != data_length as usize {
            println!("Creating apha channel...");
            for i in img.raw_pixels().iter() {
                if (image_buffer.len() + 1) % 4 == 0 {
                    image_buffer.push(255);
                }
                image_buffer.push(*i);
            }
            image_buffer.push(255);
        } else {
            image_buffer = img.raw_pixels();
        }

        let (texture, tex_future) = ImmutableImage::from_iter(
            image_buffer.iter().cloned(),
            Dimensions::Dim2d { width: xy.0, height: xy.1 },
            Format::R8G8B8A8Srgb,
            self.queue.clone()
        ).unwrap();

        texture
    }

    pub fn load_texture_from_filename(&mut self, filename: String) -> Arc<ImmutableImage<Format>> {

        if self.texture_store.contains_key(&filename.clone()) {
            println!("{} Already exists, not going to replace it.", filename.clone());
            self.texture_store.get(&filename.clone()).unwrap().clone()
        } else {
            let texture = self.get_texture_from_file(filename.clone());
            self.texture_store.insert(filename, texture.clone());
            texture
        }

    }

    // After done using this, need to call allocated vertex buffers
    pub fn draw(&mut self, drawable: &dyn Drawable) {

        match drawable.get_texture_id() {
            Some(id) => {
                self.textured_drawables
                    .entry(id)
                    .or_insert(Vec::new())
                    .extend(drawable.get_vertices().iter().map(|n|
                        Vertex2D {
                            position: [n.0, n.1],
                        }
                    ));

            },
            None => {
                let colors = drawable.get_color();

                self.colored_drawables.extend(
                    drawable.get_vertices().iter().map(|n|
                        ColoredVertex2D {
                            position: [n.0, n.1],
                            color: [colors.0, colors.1, colors.2, colors.3]
                        }
                    )
                );
            }
        }
    }


    fn get_texture(&self, texture_id: String) -> Arc<ImmutableImage<Format>> {

        if let Some(i) = self.texture_store.get(&texture_id) {
            return i.clone();
        } else {
            panic!("{} : Texture not loaded", texture_id);
        }
    }

    pub fn allocate_vertex_buffers(&mut self, device: Arc<Device>) {

        self.colored_vertex_buffer.clear();
        self.textured_vertex_buffer.clear();

        self.colored_vertex_buffer.push(
            ImmutableBuffer::from_iter(
                self.colored_drawables.iter().cloned(),
                BufferUsage::vertex_buffer(),
                self.queue.clone(),
            ).unwrap().0
        );

        for (k, v) in self.textured_drawables.drain() {
            self.textured_vertex_buffer.insert(
                k.clone(),
                ImmutableBuffer::from_iter(
                    v.iter().cloned(),
                    BufferUsage::vertex_buffer(),
                    self.queue.clone(),
                ).unwrap().0
            );
        }
    }

    fn get_solid_color_descriptor_set(&self) -> Box<dyn DescriptorSet + Send + Sync> {
        let o: Box<dyn DescriptorSet + Send + Sync> = Box::new(
            PersistentDescriptorSet::start(
                self.shader_kernels.get(&ShaderType::SOLID).unwrap().clone().get_pipeline().clone(), 0
            ).build().unwrap());
        o
    }

    fn get_textured_descriptor_set(&self, texture_id: String)
        -> Box<dyn DescriptorSet + Send + Sync> {

        let o: Box<dyn DescriptorSet + Send + Sync> = Box::new(
            PersistentDescriptorSet::start(
                self.shader_kernels.get(&ShaderType::TEXTURED).unwrap().clone().get_pipeline().clone(), 0
            )
                .add_sampled_image(self.get_texture(texture_id), self.sampler.clone()).unwrap()
                .build().unwrap());
        o
    }

    // This is the image which is written to by the write compute buffer
    // I suppose I could just have a general image set maker instead of compue... they are
    // somewhat similar
    fn get_compute_swap_descriptor_set(&mut self,
                                       device: Arc<Device>,
                                       compute_image: &ComputeImage) -> Box<dyn DescriptorSet + Send + Sync> {

        let sampler = Sampler::new(device.clone(), Filter::Linear, Filter::Linear,
                                   MipmapMode::Nearest, SamplerAddressMode::Repeat, SamplerAddressMode::Repeat,
                                   SamplerAddressMode::Repeat, 0.0, 1.0, 0.0, 0.0).unwrap();

        let o : Box<dyn DescriptorSet + Send + Sync> = Box::new(
            PersistentDescriptorSet::start(
                self.shader_kernels.get(&ShaderType::COMPUTE).clone().unwrap().clone().get_pipeline(), 0
            )
                .add_image(compute_image.clone().get_swap_buffer().clone()).unwrap()
                .build().unwrap());
        o
    }

    /*

    */
    pub fn draw_commands(&self,
                         mut command_buffer: AutoCommandBufferBuilder,
                         framebuffers: Vec<Arc<dyn FramebufferAbstract + Send + Sync>>,
                         image_num: usize) -> AutoCommandBufferBuilder {

        // Specify the color to clear the framebuffer with i.e. blue
        let clear_values = vec!(ClearValue::Float([0.0, 0.0, 1.0, 1.0]));

        let dynamic_state = DynamicState { line_width: None, viewports: None, scissors: None };

        let mut command_buffer = command_buffer.begin_render_pass(
            framebuffers[image_num].clone(), false, clear_values.clone()
        ).unwrap();


        for (shader_type, kernel) in self.shader_kernels.clone().iter() {
            match shader_type {
                ShaderType::SOLID => {
                    command_buffer = command_buffer.draw(
                        kernel.clone().get_pipeline().clone(),
                        &dynamic_state.clone(), self.colored_vertex_buffer.clone(),
                        vec![self.get_solid_color_descriptor_set()], ()
                    ).unwrap();
                },
                ShaderType::TEXTURED => {
                    command_buffer = command_buffer.draw(
                        kernel.clone().get_pipeline().clone(),
                        &dynamic_state.clone(), self.colored_vertex_buffer.clone(),
                        vec![self.get_textured_descriptor_set(String::from("funky-bird"))], ()
                    ).unwrap();
                },
                ShaderType::COMPUTE => {},
            }
        }

        command_buffer
            .end_render_pass()
            .unwrap()
    }

    /// This method is called once during initialization, then again whenever the window is resized
    pub fn window_size_dependent_setup(&self,
                                       images: &[Arc<SwapchainImage<Window>>],
    ) -> Vec<Arc<dyn FramebufferAbstract + Send + Sync>> {

        let dimensions = images[0].dimensions();

        let mut dynamic_state = DynamicState {
            line_width: None,
            viewports: Some(vec![Viewport {
                origin: [0.0, 0.0],
                dimensions: [dimensions.width() as f32, dimensions.height() as f32],
                depth_range: 0.0..1.0,
            }]),
            scissors: None
        };

        images.iter().map(|image| {
            Arc::new(
                Framebuffer::start(self.shader_kernels.get(&ShaderType::SOLID).unwrap().render_pass.clone())
                    .add(image.clone()).unwrap()
                    .build().unwrap()
            ) as Arc<dyn FramebufferAbstract + Send + Sync>
        }).collect::<Vec<_>>()
    }

}






























